import pymongo

import itertools

from pygraph.classes.digraph import digraph
from pygraph.readwrite.dot import write

import matplotlib.pyplot as plt

import networkx as nx

# on change, build interface index
# find all charms in a relation with this charm
# for each of those find any dependencies.
# if its a peer add a unit.


def pygraph_main():
    c = pymongo.Connection()
    charms = list(c['juju']['charms'].find())

    g = digraph()
    seen = set()

    for c in charms:
        name = "c.%s" % c['name']
        if name in seen:
            continue
        else:
            seen.add(name)

        nodes = [name]
        for i in itertools.chain(c["i_provides"], c["i_requires"]):
            if i in seen:
                continue
            nodes.append(i)
            seen.add(i)

        print "adding nodes", nodes
        g.add_nodes(nodes)

        for i in set(c['i_provides']):
            g.add_edge((name, i))

        for i in set(c['i_requires']):
            g.add_edge((i, name))

    with open("pygraph.dot", "w") as fh:
        fh.write(write(g))


def get_charm_graph():
    conn = pymongo.Connection()
    charms = list(conn['juju']['charms'].find())
    g = nx.digraph.DiGraph()
    seen = set()

    for c in charms:
        name = "c.%s" % c['name']
        if name in seen:
            continue
        else:
            seen.add(name)

        nodes = [name]
        for i in itertools.chain(c["i_provides"], c["i_requires"]):
            if i in seen:
                continue
            nodes.append(i)
            seen.add(i)

        g.add_nodes_from(nodes)

        for i in set(c['i_provides']):
            g.add_edge(*(name, i))

        for i in set(c['i_requires']):
            g.add_edge(*(i, name))

    conn.disconnect()
    return g


def networkx_main():
    g = get_charm_graph()
    nx.draw_random(g)  # , pos=nx.spring_layout(g))
    plt.savefig("nx-charm-graph.png")


def solve(charm_name):
    g = get_charm_graph()
    gid = "c.%s" % charm_name

    print charm_name

    clients = set()

    # Examine any provides relations
    for _, interface in g.out_edges(gid):
        for _, charm in g.out_edges(interface):
            clients.add((charm, interface))

    dependencies = set()  # Map charm id to a list of dependencies

    # Examine any depends relations
    for interface, _ in g.in_edges(gid):
        for charm, _ in g.in_edges(interface):
            if charm == gid:
                continue
            dependencies.add((charm, interface))
            #print "deps", charm, g.in_edges(charm)

    # Implicitly, peer relations are denoted by a dependency on self.
    import pprint
    print "dependencies"
    pprint.pprint(dependencies)
    print "clients"
    pprint.pprint(clients)
    print


def check_build(charm_name):
    charm_id = "c.%s" % charm_name
    g = get_charm_graph()
    results = _build_dependencies(g, charm_id)
    results = list(results)
    #print "deps", g.in_edges(charm_id)
    import pprint
    for i in results:
        print "plan"
        pprint.pprint(i)
        print


def _build_dependencies(
    graph, charm_id, requested=None, source=None, cache=None, idt=0):

    print idt * " ", "building", charm_id, requested
    idt += 1
    providers = {}
    if cache is None:
        cache = {}

    for interface, _ in graph.in_edges(charm_id):
        # skip self cycles
        if (charm_id, interface) in graph.out_edges(charm_id):
            #print "skipping", charm_id, interface, requested
            continue
        print idt * " ", "processing", interface, requested

        # Check if we have a cached plan for the providers of this
        # interface. If so append our relation for each.
        if interface in cache:
            print idt * " ", "found cached", interface
            merge_plan_relation(charm_id, interface, cache[interface])
            continue
        else:
            cache[interface] = []

        providers[interface] = []
        for provider, providing in graph.in_edges(interface):
#            if provider == charm_id:
#                print(
#                    idt * " ", "skipping", provider, providing,
#                    requested)
#                continue

            for chain in (
                _build_dependencies(
                    graph, provider,
                    requested=providing,
                    source=charm_id,
                    cache=cache,
                    idt=idt + 1)):
                print (idt + 2) * " ", "chained", provider, repr(chain)
                providers[interface].append(chain)
                cache[interface].append(chain)

            merge_plan_relation(
                provider, charm_id, interface, cache[interface])

    if not providers:
        #print idt * " ", "no providers", charm_id, requested
        yield ((charm_id, source, requested))
        raise StopIteration()

    constraints = providers.keys()
    depth = len(constraints)
    ps = set(providers.keys())
    cs = set(cache.keys())
    providers.update(cache)
    print idt * " ", "cache", cs-ps, ps-cs
    # We can get some nonsensical plans, filter. sigh.. definitely better ways.
    #import pdb; pdb.set_trace()
    #if charm_id == 'c.nova-compute':
    #    import pdb; pdb.set_trace()

    import pprint
    #print "deps", charm_id, constraints
    #print "providers", pprint.pprint(providers)
    #print
    #print "cache", pprint.pprint(cache)
    #if 'c.nova-compute' == charm_id:
    #    import pdb; pdb.set_trace()

    print
    if 0:  # charm_id == "c.nova-compute":
        print "dependencies"
        for i in providers.values():
            print i
        import pdb; pdb.set_trace()

    # Verify plans
    for plan in itertools.combinations(
        itertools.chain(*providers.values()), depth):
        required = set(constraints)
        for i in plan:
            if isinstance(i[0], basestring):
                if i[2] in required:
                    required.remove(i[2])
            elif isinstance(i[0], tuple):
                if i[0][2] in required:
                    required.remove(i[0][2])
        if required:
            continue
        yield ((charm_id, source, requested),) + plan


def find_first_relation(plan):
    # plans can be be nested
    pass


def flatten_p(plan):
    len


def merge_plan_relation(provider, charm_id, interface, plans):
    assert isinstance(plans[0], tuple)
    relation = (provider, charm_id, interface)

    for plan_idx in range(len(plans)):
        plan = plans[plan_idx]
        if provider is None:
            if isinstance(plans[0][0], basestring):
                relation = (plans[0][0], charm_id, interface)
            else:
                relation = (plans[0][0][0], charm_id, interface)
        plans[plan_idx] = (plan,) + (relation,)
        import pprint
        print "merge", pprint.pprint(plan)

if __name__ == '__main__':
    try:
        #solve('mysql')
        #solve('wordpress')
        #solve('mediawiki')

        #check_build('ceph')
        #check_build('nova-compute')
        check_build("glance")
        #check_build('mediawiki')
        #check_build('mysql')
        #solve('ceph')
        #solve('nova-compute')
        #networkx_main()
    except:
        import pdb, traceback, sys
        traceback.print_exc()
        pdb.post_mortem(sys.exc_info()[-1])
